#!/usr/bin/env python3

import dbus
import dbus.exceptions
import dbus.mainloop.glib
import dbus.service

LE_ADVERTISING_MANAGER_IFACE = 'org.bluez.LEAdvertisingManager1'
LE_ADVERTISEMENT_IFACE = 'org.bluez.LEAdvertisement1'
DBUS_PROP_IFACE = 'org.freedesktop.DBus.Properties'

class InvalidArgsException(dbus.exceptions.DBusException):
	_dbus_error_name = 'org.freedesktop.DBus.Error.InvalidArgs'


class NotSupportedException(dbus.exceptions.DBusException):
	_dbus_error_name = 'org.bluez.Error.NotSupported'


class NotPermittedException(dbus.exceptions.DBusException):
	_dbus_error_name = 'org.bluez.Error.NotPermitted'


class InvalidValueLengthException(dbus.exceptions.DBusException):
	_dbus_error_name = 'org.bluez.Error.InvalidValueLength'


class FailedException(dbus.exceptions.DBusException):
	_dbus_error_name = 'org.bluez.Error.Failed'


class Advertisement(dbus.service.Object):
	PATH_BASE = '/nelson_irrigation/twig_pilot'

	def __init__(self, bus, advertising_type):
		self.path = self.PATH_BASE
		self.bus = bus
		self.ad_type = advertising_type
		self.service_uuids = None
		self.manufacturer_data = None
		self.solicit_uuids = None
		self.service_data = None
		self.include_tx_power = None
		dbus.service.Object.__init__(self, bus, self.path)

	def get_properties(self):
		properties = dict()
		properties['Type'] = self.ad_type
		if self.service_uuids is not None:
			properties['ServiceUUIDs'] = dbus.Array(self.service_uuids, signature='s')
		if self.solicit_uuids is not None:
			properties['SolicitUUIDs'] = dbus.Array(self.solicit_uuids, signature='s')
		if self.manufacturer_data is not None:
			properties['ManufacturerData'] = dbus.Dictionary(self.manufacturer_data, signature='qay')
		if self.service_data is not None:
			properties['ServiceData'] = dbus.Dictionary(self.service_data, signature='say')
		if self.include_tx_power is not None:
			properties['IncludeTxPower'] = dbus.Boolean(self.include_tx_power)
		return {LE_ADVERTISEMENT_IFACE: properties}

	def get_path(self):
		return dbus.ObjectPath(self.path)

	def add_service_uuid(self, uuid):
		if not self.service_uuids:
			self.service_uuids = []
		self.service_uuids.append(uuid)

	def add_solicit_uuid(self, uuid):
		if not self.solicit_uuids:
			self.solicit_uuids = []
		self.solicit_uuids.append(uuid)

	def add_manufacturer_data(self, manuf_code, data):
		if not self.manufacturer_data:
			self.manufacturer_data = dict()
		self.manufacturer_data[manuf_code] = data

	def add_service_data(self, uuid, data):
		if not self.service_data:
			self.service_data = dict()
		self.service_data[uuid] = data

	@dbus.service.method(DBUS_PROP_IFACE, in_signature='s', out_signature='a{sv}')
	def GetAll(self, interface):
		if interface != LE_ADVERTISEMENT_IFACE:
			raise InvalidArgsException()
		return self.get_properties()[LE_ADVERTISEMENT_IFACE]

	@dbus.service.method(LE_ADVERTISEMENT_IFACE, in_signature='', out_signature='')
	def Release(self):
		print('{}: Released!'.format(self.path))