#!/usr/bin/env python3

import dbus
import dbus.exceptions
import dbus.mainloop.glib
import dbus.service

GATT_MANAGER_IFACE = 'org.bluez.GattManager1'
DBUS_OM_IFACE =      'org.freedesktop.DBus.ObjectManager'
DBUS_PROP_IFACE =    'org.freedesktop.DBus.Properties'
GATT_SERVICE_IFACE = 'org.bluez.GattService1'
GATT_CHRC_IFACE =    'org.bluez.GattCharacteristic1'
GATT_DESC_IFACE =    'org.bluez.GattDescriptor1'

class InvalidArgsException(dbus.exceptions.DBusException):
	_dbus_error_name = 'org.freedesktop.DBus.Error.InvalidArgs'


class NotSupportedException(dbus.exceptions.DBusException):
	_dbus_error_name = 'org.bluez.Error.NotSupported'


class NotPermittedException(dbus.exceptions.DBusException):
	_dbus_error_name = 'org.bluez.Error.NotPermitted'


class InvalidValueLengthException(dbus.exceptions.DBusException):
	_dbus_error_name = 'org.bluez.Error.InvalidValueLength'


class FailedException(dbus.exceptions.DBusException):
	_dbus_error_name = 'org.bluez.Error.Failed'

class Application(dbus.service.Object):
	def __init__(self, bus):
		self.path = '/'
		self.services = []
		dbus.service.Object.__init__(self, bus, self.path)

	def get_path(self):
		return dbus.ObjectPath(self.path)

	def add_service(self, service):
		self.services.append(service)

	@dbus.service.method(DBUS_OM_IFACE, out_signature='a{oa{sa{sv}}}')
	def GetManagedObjects(self):
		response = {}
		for service in self.services:
			response[service.get_path()] = service.get_properties()
			chrcs = service.get_characteristics()
			for chrc in chrcs:
				response[chrc.get_path()] = chrc.get_properties()
				descs = chrc.get_descriptors()
				for desc in descs:
					response[desc.get_path()] = desc.get_properties()
		return response


class Service(dbus.service.Object):
	PATH_BASE = '/com/nelson_irrigation/pilot/service'

	def __init__(self, bus, index, uuid, primary):
		self.path = self.PATH_BASE + str(index)
		self.bus = bus
		self.uuid = uuid
		self.primary = primary
		self.characteristics = []
		dbus.service.Object.__init__(self, bus, self.path)

	def get_properties(self):
		return {
			GATT_SERVICE_IFACE: {
				'UUID': self.uuid,
				'Primary': self.primary,
				'Characteristics': dbus.Array(self.get_characteristic_paths(), signature='o')
			}
		}

	def get_path(self):
		return dbus.ObjectPath(self.path)

	def add_characteristic(self, characteristic):
		self.characteristics.append(characteristic)

	def get_characteristic_paths(self):
		result = []
		for characteristic in self.characteristics:
			result.append(characteristic.get_path())
		return result

	def get_characteristics(self):
		return self.characteristics

	@dbus.service.method(DBUS_PROP_IFACE, in_signature='s', out_signature='a{sv}')
	def GetAll(self, interface):
		if interface != GATT_SERVICE_IFACE:
			raise InvalidArgsException()
		return self.get_properties[GATT_SERVICE_IFACE]


class Characteristic(dbus.service.Object):
	def __init__(self, bus, index, uuid, flags, service):
		self.path = service.path + '/char' + str(index)
		self.bus = bus
		self.uuid = uuid
		self.service = service
		self.flags = flags
		self.descriptors = []
		dbus.service.Object.__init__(self, bus, self.path)

	def get_properties(self):
		return {
			GATT_CHRC_IFACE: {
				'Service': self.service.get_path(),
				'UUID': self.uuid,
				'Flags': self.flags,
				'Descriptors': dbus.Array(self.get_descriptor_paths(), signature='o')
			}
		}

	def get_path(self):
		return dbus.ObjectPath(self.path)

	def add_descriptor(self, descriptor):
		self.descriptors.append(descriptor)

	def get_descriptor_paths(self):
		result = []
		for descriptor in self.descriptors:
			result.append(descriptor.get_path())
		return result

	def get_descriptors(self):
		return self.descriptors

	@dbus.service.method(DBUS_PROP_IFACE, in_signature='s', out_signature='a{sv}')
	def GetAll(self, interface):
		if interface != GATT_CHRC_IFACE:
			raise InvalidArgsException()
		return self.get_properties[GATT_CHRC_IFACE]

	@dbus.service.method(GATT_CHRC_IFACE, out_signature='ay')
	def ReadValue(self):
		print('Default ReadValue called, returning error')
		raise NotSupportedException()

	@dbus.service.method(GATT_CHRC_IFACE, in_signature='ay')
	def WriteValue(self, value):
		print('Default WriteValue called, returning error')
		raise NotSupportedException()

	@dbus.service.method(GATT_CHRC_IFACE)
	def StartNotify(self):
		print('Default StartNotify called, returning error')
		raise NotSupportedException()

	@dbus.service.method(GATT_CHRC_IFACE)
	def StopNotify(self):
		print('Default StopNotify called, returning error')
		raise NotSupportedException()

	@dbus.service.signal(DBUS_PROP_IFACE, signature='sa{sv}as')
	def PropertiesChanged(self, interface, changed, invalidated):
		pass


class Descriptor(dbus.service.Object):
	def __init__(self, bus, index, uuid, flags, characteristic):
		self.path = characteristic.path + '/desc' + str(index)
		self.bus = bus
		self.uuid = uuid
		self.flags = flags
		self.chrc = characteristic
		dbus.service.Object.__init__(self, bus, self.path)

	def get_properties(self):
		return {
				GATT_DESC_IFACE: {
						'Characteristic': self.chrc.get_path(),
						'UUID': self.uuid,
						'Flags': self.flags,
				}
		}

	def get_path(self):
		return dbus.ObjectPath(self.path)

	@dbus.service.method(DBUS_PROP_IFACE, in_signature='s', out_signature='a{sv}')
	def GetAll(self, interface):
		if interface != GATT_DESC_IFACE:
			raise InvalidArgsException()
		return self.get_properties[GATT_CHRC_IFACE]

	@dbus.service.method(GATT_DESC_IFACE, out_signature='ay')
	def ReadValue(self):
		print ('Default ReadValue called, returning error')
		raise NotSupportedException()

	@dbus.service.method(GATT_DESC_IFACE, in_signature='ay')
	def WriteValue(self, value):
		print('Default WriteValue called, returning error')
		raise NotSupportedException()
